/*
 * This is a simple program which changes the boot signature of a drive
 * from 0xAA55 to 0xAB55, which prevents the format warning on the Wii U.
 *
 * Author: TheDreadedAndy
 * Notes: Based on the source code to UStealth, by jayjay123
 */

#include <stdlib.h>
#include <stdio.h>

#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

/* The boot signature used by MBR */
#define MBR_SIG 0xAA55

/* The boot signature used by the Wii U */
#define WIIU_SIG 0xAB55

/* The offset of the signature in the device */
#define SIG_OFFSET 510

/* The size of the signature in the device */
#define SIG_SIZE 2

/* Helper functions */
int get_sig(int dev_fd);
void set_sig(int dev_fd, int sig);
void panic(const char *err);

/*
 * Attempts to toggle the boot signature of the given device.
 */
int main(int argc, char *argv[])
{
	if (argc != 2) {
		fprintf(stderr, "Wrong arg count.\n");
		fprintf(stderr, "Usage: uustealth <device file>\n");
		return -1;
	}

	const char *dev = argv[1];
	int dev_fd = open(dev, O_RDWR);
	if (dev_fd < 0) {
		fprintf(stderr, "Failed to open specified device.\n");
		fprintf(stderr, "Try running the program as root.\n");
		return -1;
	}

	int sig = get_sig(dev_fd);
	if (sig == MBR_SIG) {
		set_sig(dev_fd, WIIU_SIG);
		printf("Your drive is now hidden.\n"
		       "Standard devices may ask to format it.\n");
	} else if (sig == WIIU_SIG) {
		set_sig(dev_fd, MBR_SIG);
		printf("Your drive is now visible.\n"
		       "Your Wii U will ask to to format it on boot.\n");
	} else {
		fprintf(stderr, "Unknown boot signature.\n"
		        "Is the drive not MBR?\n");
		return -1;
	}

	return 0;
}

/*
 * Gets the current boot signature of the given device.
 */
int get_sig(int dev_fd)
{
	if (dev_fd < 0) {
		panic("Can't get the boot sig for an invalid device.\n");
	}

	if (lseek(dev_fd, SIG_OFFSET, SEEK_SET) < 0) {
		panic("Cannot seek with given device.\n");
	}

	unsigned char sig[SIG_SIZE];
	if (SIG_SIZE != read(dev_fd, sig, SIG_SIZE)) {
		panic("Cannot read signature.\n");
	}

	return sig[0] | (sig[1] << 8);
}

/*
 * Sets the boot signature of the given device.
 */
void set_sig(int dev_fd, int sig)
{
	if (dev_fd < 0) {
		panic("Can't set the boot sig for an invalid device.\n");
	}

	if ((sig != WIIU_SIG) && (sig != MBR_SIG)) {
		panic("Can't set sig to bad value.\n");
	}

	if (lseek(dev_fd, SIG_OFFSET, SEEK_SET) < 0) {
		panic("Cannot seek with the given device.\n");
	}

	unsigned char buf[SIG_SIZE];
	buf[0] = sig;
	buf[1] = sig >> 8;
	if (SIG_SIZE != write(dev_fd, buf, SIG_SIZE)) {
		panic("Failed to update drive signature.\n");
	}
}

/*
 * Aborts the application and displays an error.
 */
void panic(const char *err) {
	if (err == NULL) { err = "Panic called with bad err.\n"; }

	fprintf(stderr, err);
	abort();
}
